import 'dart:convert';

import 'package:archive/archive.dart';
import 'package:epub/epub.dart';
import 'package:epub/src/ref_entities/epub_byte_content_file_ref.dart';
import 'package:epub/src/ref_entities/epub_content_file_ref.dart';
import 'package:epub/src/ref_entities/epub_content_ref.dart';
import 'package:epub/src/ref_entities/epub_text_content_file_ref.dart';
import 'package:epub/src/schema/opf/epub_version.dart';
import 'package:epub/src/utils/zip_path_utils.dart';
import 'package:xml/xml.dart';

class EpubReferencer {
  static const _container_file =
      '<?xml version="1.0"?><container version="1.0" xmlns="urn:oasis:names:tc:opendocument:xmlns:container"><rootfiles><rootfile full-path="OEBPS/content.opf" media-type="application/oebps-package+xml"/></rootfiles></container>';

  static EpubBookRef reference(EpubBook book) {
    var a = new Archive();
    a.addFile(new ArchiveFile.noCompress(
        "metadata", 20, utf8.encode("application/epub+zip")));
    var ref = new EpubBookRef(a);

    a.addFile(new ArchiveFile("META-INF/container.xml", _container_file.length,
        utf8.encode(_container_file)));

    ref.Title = book.Title;
    ref.Author = book.Author;
    ref.AuthorList = book.AuthorList;
    ref.FilePath = book.FilePath;
    ref.Schema = book.Schema;
    ref.Content = referenceContent(book.Content, a, ref);

    return ref;
  }

  static EpubContentRef referenceContent(
      EpubContent content, Archive a, EpubBookRef eref) {
    var ref = new EpubContentRef();

    ref.AllFiles = new Map();
    ref.Html = new Map();
    ref.Css = new Map();
    ref.Images = new Map();
    ref.Fonts = new Map();

    content.AllFiles.forEach((name, file) {
      EpubContentFileRef fileref;
      if (file is EpubByteContentFile) {
        fileref = new EpubByteContentFileRef(eref);
      } else if (file is EpubTextContentFile) {
        fileref = new EpubTextContentFileRef(eref);
      }
      fileref.FileName = file.FileName;
      fileref.ContentType = file.ContentType;
      fileref.ContentMimeType = file.ContentMimeType;

      List<int> content;

      if (file is EpubByteContentFile) {
        content = file.Content;
      } else if (file is EpubTextContentFile) {
        content = utf8.encode(file.Content);
      }

      a.addFile(new ArchiveFile(
          ZipPathUtils.combine(eref.Schema.ContentDirectoryPath, name),
          content.length,
          content));
      ref.AllFiles[name] = fileref;
    });

    content.Html.forEach((name, _) {
      ref.Html[name] = ref.AllFiles[name];
    });

    content.Css.forEach((name, _) {
      ref.Css[name] = ref.AllFiles[name];
    });

    content.Images.forEach((name, _) {
      ref.Images[name] = ref.AllFiles[name];
    });

    content.Fonts.forEach((name, _) {
      ref.Fonts[name] = ref.AllFiles[name];
    });

    var contentopf = getContentOPF(eref);
    a.addFile(new ArchiveFile(ZipPathUtils.combine(eref.Schema.ContentDirectoryPath, "content.opf"), contentopf.length, utf8.encode(contentopf)));

    return ref;
  }

  // todo probably not complete
  static String getContentOPF(EpubBookRef book) {
    const dc_namespace = "http://purl.org/dc/elements/1.1/";
    const opf_namespace = "http://www.idpf.org/2007/opf";
    var builder = new XmlBuilder();
    builder.processing('xml', 'version="1.0"');

    builder.element("package", attributes: {
      "version":
          book.Schema.Package.Version == EpubVersion.Epub2 ? "2.0" : "3.0",
      "unique-identifier": "etextno",
    }, nest: () {
      builder.namespace("http://www.idpf.org/2007/opf");
      builder
        ..element("metadata",
            namespaces: {opf_namespace: "opf", dc_namespace: "dc"}, nest: () {
          var meta = book.Schema.Package.Metadata;

          meta
            ..Titles.forEach((item) =>
                builder.element("title", nest: item, namespace: dc_namespace))
            ..Creators.forEach((item) =>
                builder.element("creator", namespace: dc_namespace, nest: () {
                  if (item.Role != null)
                    builder.attribute("role", item.Role,
                        namespace: opf_namespace);
                  if (item.FileAs != null)
                    builder.attribute("file-as", item.FileAs,
                        namespace: opf_namespace);

                  builder.text(item.Creator);
                }))
            ..Subjects.forEach((item) =>
                builder.element("subject", namespace: dc_namespace, nest: item))
            ..Publishers.forEach((item) => builder.element("publisher",
                namespace: dc_namespace, nest: item))
            ..Contributors.forEach((item) => builder
                    .element("contributor", namespace: dc_namespace, nest: () {
                  if (item.Role != null)
                    builder.attribute("role", item.Role,
                        namespace: opf_namespace);
                  if (item.FileAs != null)
                    builder.attribute("file-as", item.FileAs,
                        namespace: opf_namespace);

                  builder.text(item.Contributor);
                }))
            ..Dates.forEach((date) =>
                builder.element("date", namespace: dc_namespace, nest: () {
                  if (date.Event != null)
                    builder.attribute("event", date.Event,
                        namespace: opf_namespace);

                  builder.text(date.Date);
                }))
            ..Types.forEach((type) =>
                builder.element("type", namespace: dc_namespace, nest: type))
            ..Formats
                .forEach((format) => builder.element("format", namespace: dc_namespace, nest: format))
            ..Identifiers.forEach((id) => builder.element("identifier", namespace: dc_namespace, nest: () {
                  if (id.Id != null) builder.attribute("id", id.Id);
                  if (id.Scheme != null)
                    builder.attribute("scheme", id.Scheme,
                        namespace: opf_namespace);
                  builder.text(id.Identifier);
                }))
            ..Sources.forEach((item) => builder.element("source", namespace: dc_namespace, nest: item))
            ..Languages.forEach((item) => builder.element("language", namespace: dc_namespace, nest: item))
            ..Relations.forEach((item) => builder.element("relation", namespace: dc_namespace, nest: item))
            ..Coverages.forEach((item) => builder.element("coverage", namespace: dc_namespace, nest: item))
            ..Rights.forEach((item) => builder.element("rights", namespace: dc_namespace, nest: item))
            ..MetaItems.forEach((metaitem) => builder.element("meta", nest: () {
                  if (book.Schema.Package.Version == EpubVersion.Epub2) {
                    if (metaitem.Name != null)
                      builder.attribute("name", metaitem.Name);
                    if (metaitem.Content != null)
                      builder.attribute("content", metaitem.Content);
                  } else if (book.Schema.Package.Version == EpubVersion.Epub3) {
                    if (metaitem.Id != null)
                      builder.attribute("id", metaitem.Id);
                    if (metaitem.Refines != null)
                      builder.attribute("refines", metaitem.Refines);
                    if (metaitem.Property != null)
                      builder.attribute("property", metaitem.Property);
                    if (metaitem.Scheme != null)
                      builder.attribute("scheme", metaitem.Scheme);
                  }
                }));

          if (meta.Description != null)
            builder.element("description",
                namespace: dc_namespace, nest: meta.Description);
        })
        ..element("manifest", nest: () {
          book.Schema.Package.Manifest.Items.forEach((item) {
            builder.element("item", nest: () {
              builder
                ..attribute("id", item.Id)
                ..attribute("href", item.Href)
                ..attribute("media-type", item.MediaType);
            });
          });
        })
        ..element("spine",
            attributes: {"toc": book.Schema.Package.Spine.TableOfContents},
            nest: () {
          book.Schema.Package.Spine.Items.forEach((spineitem) => builder
              .element("itemref", attributes: {"idref": spineitem.IdRef}));
        });
    });

    return builder.build().toXmlString(pretty: true);
  }
}
